#include <GL/glew.h>
#include <GL/wglew.h>
#include <FK/FK.h>
#include "FKUT/TrackBall.h"
#include "FKUT/GuideObject.h"
#include "GLSLModel.h"

/*
** _[^[Qbg̃Xg
*/
static const GLenum bufs[] = {
	GL_COLOR_ATTACHMENT0_EXT, //   J[obt@
	GL_DEPTH_ATTACHMENT_EXT, //   fvXeNX`
};

class MyWindow : public fk_Window {
private:
	GLuint		fboID, texID, depthTexID, rendID;
	GLuint		vpID, fpID, shID;
	int			bufW, bufH;
	int			size;
	float		weight, pintCenter, pintDist;

	void printShaderInfoLog(GLuint shader)
	{
		GLsizei bufSize;

		/* VF[_̃RpC̃O̒擾 */
		glGetShaderiv(shader, GL_INFO_LOG_LENGTH , &bufSize);

		if(bufSize > 1) {
			GLchar *infoLog = (GLchar *)malloc(bufSize);

			if(infoLog != NULL) {
				GLsizei length;

				/* VF[_̃RpC̃O̓e擾 */
				glGetShaderInfoLog(shader, bufSize, &length, infoLog);
				fprintf(stderr, "InfoLog:\n%s\n\n", infoLog);
				free(infoLog);
			} else {
				fprintf(stderr, "Could not allocate InfoLog buffer.\n");
			}
		}
	};

	void loadShaderFile(GLuint id, const char *argFileName)
	{
		// t@C̒gSēǂݍ
		std::ifstream inputFile1(argFileName);
		std::istreambuf_iterator<char> vdataBegin(inputFile1);
		std::istreambuf_iterator<char> vdataEnd;
		std::string fileData1(vdataBegin,vdataEnd);
		const char *pFile = fileData1.c_str();

		glShaderSource(id, 1, &pFile, NULL);
		glCompileShader(id);
	};

	void initFBO(void)
	{
		// J[obt@p̃eNX`pӂ
		glGenTextures(1, &texID);
		glBindTexture(GL_TEXTURE_2D, texID);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, bufW, bufH, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glBindTexture(GL_TEXTURE_2D, 0);

		// fvXeNX`p̃eNX`pӂ
		glGenTextures(1, &depthTexID);
		glBindTexture(GL_TEXTURE_2D, depthTexID);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, bufW, bufH, 0, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, 0);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glBindTexture(GL_TEXTURE_2D, 0);

		// fvXobt@p̃_[obt@pӂ
		/*
		glGenRenderbuffersEXT(1, &rendID);
		glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, rendID);
		glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT, bufW, bufH);
		glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, 0);
		*/

		// t[obt@IuWFNg쐬
		glGenFramebuffersEXT(1, &fboID);
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fboID);

		// t[obt@IuWFNgɃJ[obt@ƂăeNX`
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, texID, 0);
		// t[obt@IuWFNgɃfvXeNX`
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_TEXTURE_2D, depthTexID, 0);

		// t[obt@IuWFNgɃfvXobt@Ƃă_[obt@
		//glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, rendID);

		// t[obt@IuWFNǧ
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	};

	void initShader(void)
	{
		vpID = glCreateShader(GL_VERTEX_SHADER);
		loadShaderFile(vpID, "shader/fbo_vp.glsl");
		printShaderInfoLog(vpID);

		fpID = glCreateShader(GL_FRAGMENT_SHADER);
		loadShaderFile(fpID, "shader/fbo_fp_gau.glsl");
		printShaderInfoLog(fpID);

		shID = glCreateProgram();
		glAttachShader(shID, vpID);
		glAttachShader(shID, fpID);

		GLint	linked = 0;
		glLinkProgram(shID);
		glGetProgramiv(shID, GL_LINK_STATUS, &linked);
		if(linked == GL_FALSE) {
			std::cerr << "Link Err.\n";
		}

		return;
	};

public:
	MyWindow(int x = 0, int y = 0, int w = 300, int h = 300, std::string name = "")
		: fk_Window(x, y, w, h, name), size(5), weight(1.0), pintCenter(0.5f), pintDist(0.01f)
	{
		bufW = w; bufH = h;
		fboID = texID = rendID = 0;
		vpID = fpID = shID = 0;
	};

	~MyWindow()
	{
		glDeleteShader(vpID);
		glDeleteShader(fpID);
		glDeleteProgram(shID);
	};

	void preInit(void)
	{
		// GLEW ̏
#ifdef WIN32
		GLenum err = glewInit();
		if (err != GLEW_OK) {
			fprintf(stderr, "Error: %s\n", glewGetErrorString(err));
			exit(1);
		}
		wglSwapIntervalEXT(0);
#endif

		initFBO();
		initShader();
  
		return;
	};

	void postInit(void)
	{
		return;
	};

	void preDraw(void)
	{
		// t[obt@IuWFNg
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fboID);
		// _[^[Qbgw肷
		glDrawBuffers(sizeof bufs / sizeof bufs[0], bufs);
		// FKł̕`eFBOɃXCb`
		return;
	};

	void postDraw(void)
	{
		// _[^[Qbgɖ߂
		glDrawBuffer(GL_BACK);
		// t[obt@IuWFNg
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

		// ϊsPʍsɂ
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();

		// fr[ϊsPʍsɂ
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();

		// AetƉBʏ͍sȂ
		glDisable(GL_LIGHTING);
		glDisable(GL_DEPTH_TEST);

		// eNX`}bsOLɂ
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, texID);
		glActiveTexture(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_2D, depthTexID);
		//glEnable(GL_TEXTURE_2D);

		glUseProgram(shID);
		glUniform1i(glGetUniformLocation(shID, "size"), size);
		glUniform1f(glGetUniformLocation(shID, "weight0"), weight);
		glUniform1f(glGetUniformLocation(shID, "pintCenter"), pintCenter);
		glUniform1f(glGetUniformLocation(shID, "pintDist"), pintDist);
		glUniform1i(glGetUniformLocation(shID, "tex0"), 0);
		glUniform1i(glGetUniformLocation(shID, "depthTex"), 1);

		// ``
		glColor3d(1.0, 1.0, 1.0);
		glBegin(GL_TRIANGLE_FAN);
		glTexCoord2d(0.0, 0.0);
		glVertex2d(-1.0, -1.0);
		glTexCoord2d(1.0, 0.0);
		glVertex2d( 1.0, -1.0);
		glTexCoord2d(1.0, 1.0);
		glVertex2d( 1.0,  1.0);
		glTexCoord2d(0.0, 1.0);
		glVertex2d(-1.0,  1.0);
		glEnd();

		glUseProgram(0);

		// eNX`}bsO𖳌ɂ
		//glDisable(GL_TEXTURE_2D);
		glActiveTexture(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_2D, 0);
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, 0);

		glEnable(GL_DEPTH_TEST);
		glEnable(GL_LIGHTING);

		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();

		glFlush();

		return;
	};

	void setBokeguai(int argSize, float argWeight, float argPintCenter, float argPintDist)
	{
		size = argSize;
		weight = argWeight;
		pintCenter = argPintCenter;
		pintDist = argPintDist;
	};
};

float distToDepth(float argDist, float argNear, float argFar)
{
	const float a = argFar / (argFar - argNear);
	const float b = -(argFar * argNear) / (argFar - argNear);
	float ret = (a * argDist + b) / argDist;
	return ret;
}

int main(int argc, char *argv[])
{
	Fl_Window		mainWindow(960, 540, "FK TEST");
	fk_Model		camera, lightModel;
	fk_Light		light;
	fk_Scene		scene;
	MyWindow		window(0, 0, 960, 540);

	fkut_GuideObject	guide;
	fk_TrackBall		tb(&window, &camera);

	fk_IFSTexture	ifsShape;
	fk_Sphere		sph(8, 7.0);

	GLSLModel		modelDef, ifsModelDef;
	GLSLModel		sphShockModel;

	fk_Block		blk(100.0, 10.0, 1.0);
	fk_Model		blkModel;

	fk_System::setcwd();
	mainWindow.end();
	fk_InitMaterial();
	fk_SetErrorMode(FK_ERR_BROWSER_INTERACTIVE);

	// Ɩ̐ݒ
	lightModel.setShape(&light);
	lightModel.setMaterial(White);
	lightModel.getMaterial()->setAmbDiff(1.0f, 1.0f, 1.0f);
	lightModel.getMaterial()->setSpecular(1.0f, 1.0f, 1.0f);
	lightModel.glMoveTo(0.0, 0.0, 0.0);
	lightModel.glFocus(-1.0, -1.0, -1.0);

	// f̓ǂݍ
	if(!ifsShape.readBMP("mqo/00tex_master.BMP")) fl_alert("tex load err");
	if(!ifsShape.readMQOFile("mqo/meka.mqo", "body01")) fl_alert("ifs load err");
	ifsShape.setTexRendMode(FK_TEX_REND_SMOOTH);

	blkModel.setShape(&blk);
	blkModel.setMaterial(Red);
	blkModel.setParent(&camera);

	// effBXvCXgɓo^
	scene.setBlendStatus(true);
	scene.setBGColor(0.5f, 0.5f, 0.5f);
	scene.entryCamera(&camera);
	scene.entryModel(&lightModel);
	scene.entryModel(&blkModel);
	guide.entryScene(&scene);

	fk_Perspective	pers;
	pers.setFar(600.0);
	scene.setProjection(&pers);

	// EBhEփfBXvCXgo^
	window.setScene(&scene);

	// _̈ʒuƎpݒ
	camera.glMoveTo(0.0, 0.0, 100.0);
	camera.glFocus(0.0, 0.0, 0.0);
	camera.glUpvec(0.0, 1.0, 0.0);

	mainWindow.show();
	window.show();
	Fl::check();

	modelDef.setShape(&sph);
	modelDef.setMaterial(Yellow);
	modelDef.setSmoothMode(true);
	modelDef.glMoveTo(-20.0, 0.0, 0.0);
	modelDef.initShader(0, 0);
	scene.entryModel(&modelDef);

	ifsModelDef.setShape(&ifsShape);
	ifsModelDef.setMaterial(White);
	ifsModelDef.setSmoothMode(true);
	ifsModelDef.glMoveTo(20.0, 0.0, 0.0);
	ifsModelDef.initShader(0, 1);
	scene.entryModel(&ifsModelDef);

	sphShockModel.setShape(&sph);
	sphShockModel.setSmoothMode(true);
	sphShockModel.setScale(0.01);
	sphShockModel.initShader(0, 2);
	scene.entryModel(&sphShockModel);

	int size = 5;
	float weight = 1.0f, pintCenter = 1.0f, pintDist = 0.01f;
	bool pre1 = false, pre2 = false;

	while(true) {
		if(mainWindow.visible() == 0) {
			if(Fl::wait() == 0) {
				break;
			} else {
				continue;
			}
		}

		if(window.drawWindow() == 0) break;
		if(Fl::check() == 0) break;
		if(window.winOpenStatus() == false) continue;

		lightModel.glRotateWithVec(0.0, 0.0, 0.0, fk_Y, 0.01);

		if(window.getKeyStatus(' ', false)) {
			sphShockModel.setScale(sphShockModel.getScale()+0.1);
		}
		sphShockModel.glFocus(-camera.getPosition());

		if(window.getKeyStatus('1') & !pre1) {
			if(size > 1) --size;
			cout << "Size: " << size << endl;
		}
		if(window.getKeyStatus('2') & !pre2) {
			++size;
			cout << "Size: " << size << endl;
		}
		if(window.getKeyStatus('3')){
			if(weight > 0.0) weight -= 0.1f;
			cout << "Weight: " << weight << endl;
		}
		if(window.getKeyStatus('4')){
			weight += 0.1f;
			cout << "Weight: " << weight << endl;
		}
		if(window.getKeyStatus('5')){
			if(pintCenter > 1.0) pintCenter -= 1.0;
			blkModel.glMoveTo(0.0, 0.0, -pintCenter);
			cout << "PintCenter: " << pintCenter << endl;
		}
		if(window.getKeyStatus('6')){
			if(pintCenter < 600.0) pintCenter += 1.0;
			blkModel.glMoveTo(0.0, 0.0, -pintCenter);
			cout << "PintCenter: " << pintCenter << endl;
		}
		if(window.getKeyStatus('7')){
			if(pintDist > 0.01f) pintDist -= 0.01f;
			cout << "PintDist: " << pintDist << endl;
		}
		if(window.getKeyStatus('8')){
			if(pintDist < 0.5f) pintDist += 0.01f;
			cout << "PintDist: " << pintDist << endl;
		}
		window.setBokeguai(size, weight, distToDepth(pintCenter, 1.0f, 600.0f), pintDist);
		pre1 = window.getKeyStatus('1');
		pre2 = window.getKeyStatus('2');

		tb.update();
	}

	return 0;
}
